% Dixon / eigenvalue solver for 3 point calibrated pose problem.
% For each pair of input points we have a cosine rule constraint
%
%	Pij(ui,uj) = ui^2+uj^2-2*cos(theta_ij)*ui*uj - dij^2 = 0
%
% where ui,uj are the camera-point distances, dij the distance between
% the corresponding 3D points, and theta_ij the angle between the
% direction vectors from the camera to the 3D points, as measured by the
% calibrated camera. For convenience we define 
%
%	Cij = 2*cos(theta_ij)  and  Dij = dij^2
%
% The matrices below were generated by Maple using B. Mourrain's
% `duality' package. For convenience, we first changed variables to
%
%	y1 = 1/u1^2, y2 = u2/u1, y3 = u3/u1
%
% to get the system 
%
%	P12' =  1   + y2^2 - C12*y2    - D12*y1 = 0
%	P13' =  1   + y3^2 - C13*y3    - D13*y1 = 0
%	P23' = y2^2 + y3^2 - C23*y2*y3 - D23*y1 = 0
%
% Eliminating y2,y3 using the Bezout-Cayley-Dixon resultant gives a 
% 5x5 matrix R(y1) = R0 + R1*y1, whose determinant vanishes when there
% is a solution. We solve this by the eigensystem (inv(R0)*R1)*v = v/y1
% R1 is always singular, so there is a discardable eigenvalue 1/y1=0,
% and 4 useful ones. y2,y3 are found from the corresponding
% eigenvectors, which are proportional to [1,y3,y2*y3,y2,y2^2].

function u = pose3dixon(C12,D12, C13,D13, C23,D23)
   R0 = zeros(5);
   R1 = zeros(5);
   t1 = C12*C23;
   t3 = C12*C13;
   t4 = -t3+C23;
   t5 = C23*C13;
   R0(1,2) = C12;
   R0(1,3) = -2;
   R0(1,4) = -t1+C13;
   R0(1,5) = C23;
   R0(2,1) = C13;
   R0(2,2) = -2;
   R0(2,3) = C12;
   R0(2,4) = t4;
   R0(3,1) = C12-t5;
   R0(3,2) = t4;
   R0(3,3) = C13;
   R0(3,4) = -2+t3*C23;
   R0(3,5) = -t5;
   R0(4,1) = C23;
   R0(4,4) = -t1;
   R0(4,5) = C23;
   R0(5,1) = -2;
   R0(5,2) = C13;
   R0(5,3) = -C23;
   R0(5,4) = C12;

   t1 = D23*C12;
   t5 = D13-D23+D12;
   t7 = D13*C12;
   t9 = D12*C13;
   t11 = D13*C23;
   t16 = D12*C23;
   R1(1,1) = -t1*C13;
   R1(1,2) = -C12*(D13-D23);
   R1(1,3) = t5;
   R1(1,4) = D23*C13+t7*C23-t9;
   R1(1,5) = -t11;
   R1(2,1) = -C13*(D12-D23);
   R1(2,2) = t5;
   R1(2,4) = -t11;
   R1(3,1) = t1-t7+t9*C23;
   R1(3,2) = -t16;
   R1(3,4) = t5;
   R1(4,1) = -t16;
   R1(5,1) = t5;

   R = -R0 \ R1;
   [V,E] = eig(R);
   E = diag(E);
   u = zeros(0,3);
   emin = min(abs(real(E)));
   for i = 1:5
      if ~imag(E(i)) & real(E(i)) > emin 
	 e = real(E(i));
	 v = real(V(:,i))';
	 if max(abs(v([2,4]))) >= max(abs(v([3,5])))
	    u = [u; sqrt(e)*[1,v([4,2])/v(1)] ];
	 else
	    u = [u; sqrt(e)*[1,v([5,3])/v(4)] ];	    
	 end;
      end;
   end;

%end;
